from .base import BaseBuilder
import logging

logger = logging.getLogger(__name__)

sitemap_str = """<?xml version="1.0" encoding="UTF-8"?>
<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" 
  xmlns:image="http://www.google.com/schemas/sitemap-image/1.1" 
  xmlns:video="http://www.google.com/schemas/sitemap-video/1.1">
{urls}
</urlset>
"""

url_str = """  <url>
    <loc>{url}</loc>
  </url>"""

class SitemapBuilder(BaseBuilder):

    def build_sitemap(self, language, links):
        urls = set()
        site_url = self.site.url(language)
        link_attrs = self.site.link_attrs[language]
        for name in links:
          attrs = link_attrs.get(name, {})
          if not attrs.get('sitemap', True):
            logger.info("Excluding '{}' from sitemap...".format(name))
            continue
          url = self.site.full_href(language, name)
          urls.add(url_str.format(url=url))
        sitemap = sitemap_str.format(urls="\n".join(list(urls)))
        dst = self.site.get_dst('sitemap', language, extension='xml')
        self.site.write(sitemap, dst)

    def build(self):
        for language, links in self.site.links.items():
            self.build_sitemap(language, links)


import re
import os
import logging
import traceback

from .helpers.hash import hash
from .helpers.yaml import yaml
from .helpers.translate import translate
from .helpers.languages import get_source_and_target_languages
from .helpers.serialize import serialize_text, deserialize_text

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("")

def translate_data(token, data, source_language, target_language, parent=None):
    count = 0
    for key, value in data.items():
        if isinstance(value, dict):
            if source_language in value:
                source_text = value[source_language]
                # we strip multiple newlines and other stuff, as it confuses DeepL
                source_text = re.sub(r"[\s\n\t]+", " ", source_text)
                value[source_language] = source_text
                source_text = serialize_text(source_text)
                source_hash = hash(source_text)
                full_key = key
                if parent:
                    full_key = f'{parent}.{key}'
                if not '_t' in value:
                    value['_t'] = {}
                _t = value['_t']
                _i = value.get('_i', set())
                # we ignore the target language if it's listed in the '_i' field
                if target_language in _i:
                    logger.debug(f"Ignoring translation for key '{full_key}' to language '{target_language}'...")
                    continue
                # we check if we already translated this...
                if target_language in _t and _t[target_language] == source_hash:
                    logger.debug(f"Hash still ok for key '{full_key}', skipping...")
                    continue
                # we remember the hash of the source text, so we don't translate
                # it again un-necessarily...
                _t[target_language] = source_hash
                print(f"Translating '{full_key}' from '{source_language}' to '{target_language}'")
                count += len(source_text)
                value[target_language] = deserialize_text(translate(source_text, source_language, target_language, token))
            else:
                try:
                    count += translate_data(token, value, source_language, target_language, parent=key if parent is None else parent+'.'+key)
                except:
                    logger.error(traceback.format_exc())
                    continue
    return count

def translate_file(token, path, source_language, target_languages):
    count = 0
    with open(path) as input_file:
        data = yaml.load(input_file.read(), Loader=yaml.FullLoader)
    for target_language in target_languages:
        count += translate_data(token, data, source_language, target_language)
    with open(path, "w") as output_file:
        output_file.write(yaml.dump(data, indent=2, sort_keys=True))
    return count

def translate_strings(token, src_path, clean=False):

    translations_path = os.path.join(src_path, "translations")
    source_language, target_languages = get_source_and_target_languages(src_path)

    logger.info(f"Translating from '{source_language}' to '{', '.join(target_languages)}'...")

    for root, dirs, files in os.walk(translations_path):
        for filename in files:
            if filename.endswith(".yml"):
                path = os.path.join(root, filename)
                count = translate_file(token, path, source_language, target_languages)
                if count:
                    print(f"Translated {count} characters in file {path}")



from .base import BaseBuilder

import logging
import shutil
import os

logger = logging.getLogger(__name__)

class StaticFilesBuilder(BaseBuilder):

    def __init__(self, site):
        super().__init__(site)
        self._static_paths = None
        self.providers = {
            'static_file' : self.get_path,
            'source_path': self.get_source_path
        }

    def postprocess(self):
        self.copy_static_files()

    def copy_static_files(self):
        dirs = self.get_static_paths()
        for dir in dirs:
            for path, dirs, files in os.walk(dir):
                relpath = os.path.relpath(path, dir)
                for dirname in dirs:
                    dest_path = os.path.join(self.site.build_path, self.site.static_path, relpath, dirname)
                    if os.path.exists(dest_path) and not os.path.isdir(dest_path):
                        shutil.rmtree(dest_path)
                    elif not os.path.exists(dest_path):
                        os.makedirs(dest_path)
                for filename in files:
                    src_path = os.path.join(path, filename)
                    dest_path = os.path.join(self.site.build_path, 
                                             self.site.static_path,
                                             relpath,
                                             filename)
                    if not os.path.exists(dest_path) or \
                      os.stat(dest_path).st_mtime < os.stat(src_path).st_mtime:
                        shutil.copy(src_path, dest_path)

    def resolve_path(self, path):
        dirs = self.get_static_paths()
        for dir in dirs:
            full_path = os.path.join(dir, path)
            if os.path.exists(full_path):
                return full_path
        logger.error("Path '{}' not found!".format(path))

    def get_source_path(self, path):
        dirs = self.get_static_paths()
        for dir in dirs:
            full_path = os.path.join(dir, path)
            if os.path.exists(full_path):
                return full_path


    def get_path(self, path):
        full_path = self.resolve_path(path)
        if not full_path:
            return
        return os.path.join(self.site.path, self.site.static_path, path)

    def get_static_paths(self):
        if self._static_paths is not None:
            return self._static_paths
        paths = [os.path.join(self.site.src_path, 'static')]
        for language in self.site.config.get('languages', {}):
            static_path = os.path.join(self.site.src_path, '{}/static'.format(language))
            if os.path.exists(static_path):
                paths.append(static_path)
        self._static_paths = paths
        return paths




import re
import base64

def serialize_code(match):
    """
    We retain the original value inside because it can provide meaningful context
    to the translation API. However, we ignore the translated value.
    """
    return f"<code>{match.group(1)}</code>"

def serialize_ignore(match):
    # brackets don't match
    if not len(match.group(1)) == len(match.group(3)):
        return match.group(0)
    bi = base64.urlsafe_b64encode(match.group(0).encode("utf-8")).decode("ascii")
    return f"<ignore v=\"{bi}\" />"

def deserialize_ignore(match):
    bs = base64.urlsafe_b64decode(match.group(1)).decode("utf-8")
    return bs

def serialize_md_link(match):
    if len(match.groups()) == 2:
        bi = base64.urlsafe_b64encode(match.group(2).encode("utf-8")).decode("ascii")
        return f"<md-link href=\"{bi}\">{match.group(1)}</md-link>"
    else:
        return f"<md-link>{match.group(1)}</md-link>"

def deserialize_md_link(match):
    if len(match.groups()) == 2:
        bs = base64.urlsafe_b64decode(match.group(1)).decode("utf-8")
        return f"[{match.group(2)}]({bs})"
    else:
        return f"[{match.group(1)}]"
    return bs


def deserialize_text(text):
    """
    Undoes what the serialization did below.
    """

    # sometimes space after HTML tags gets removed, we fix that here
    text = re.sub(r"</([a-zA-Z\-]+)>([^$\.\;\<\n\s])", "</\\1> \\2", text)

    text = re.sub(r"<md-heading\s+v=\"(.*?)\"\s*>(.*?)</md-heading>", "\\1 \\2", text)
    text = re.sub(r"<md-list\s+v=\"(.*?)\"\s*>(.*?)</md-list>", "\\1 \\2", text)
    text = re.sub(r"<md-it>(.*?)</md-it>", "*\\1*", text)
    text = re.sub(r"<md-strong>(.*?)</md-strong>", "**\\1**", text)
    text = re.sub(r"<md-strong-it>(.*?)</md-strong-it>", "***\\1***", text)
    text = re.sub(r"<md-link\s+href=\"(.*?)\"\s*>(.*?)</md-link>", deserialize_md_link, text)
    text = re.sub(r"<md-link>(.*?)</md-link>", deserialize_md_link, text)
    # ignore needs to be deserialized last, as it can occur encoded e.g. in a
    # markdown link that was converted itself...
    text = re.sub(r"<ignore\s+v=\"(.*?)\"\s*/>", deserialize_ignore, text)
    text = text.replace("&amp;", "&")
    return text

def serialize_text(text):
    """
    Serializes text so that DeepL can effectively translate it.

    Currently, this replaces both common Markdown directives and some Jinja
    directives that might get in the way of a proper translation.
    """
    # ignore Jinja filters and JS/Python styled string formatting directives
    replaced_text = ""
    remaining_text = text
    while remaining_text:
        # this is a really simple HTML tag detector. It assumes that no
        # unescaped '<' characters occur in the source code, which is usually
        # a reasonable assumption. This is not a standard-compliant HTML parser!
        match = re.match(r"^([^<]*?)((?!\\)<.*?(?!\\)>)", remaining_text)
        if match:
            plaintext = match.group(1)
            html = match.group(2)
            if plaintext:
                replaced_text += serialize_plaintext(plaintext)
            replaced_text += html
            remaining_text = remaining_text[len(match.group(0)):]
        else:
            replaced_text += serialize_plaintext(remaining_text)
            break
    return replaced_text

def serialize_plaintext(text):
    # we ignore everything inside backticks
    text = re.sub(r"`(.*?)`", serialize_code, text)
    # we ignore everything inside unescaped brackets ({...})
    text = re.sub(r"((?:(?!\\)\{)+)(.*?)((?:(?!\\)\})+)", serialize_ignore, text)
    # we replace Markdown headings
    text = re.sub(r"^(\#+)\s*(.+?)(\n|$)", "<md-heading v=\"\\1\">\\2</md-heading>\\3", text, re.MULTILINE)
    # we replace Markdown list elements
    text = re.sub(r"^(\s*\*|\-|\d+\.)\s+(.+?)(\n|$)", "<md-list v=\"\\1\">\\2</md-list>\\3", text, re.MULTILINE)
    # we replace strong, italicized and strong italicized text
    text = re.sub(r"(?:(?![\\])\*){3}([^\*\n]+)(?:(?![\\])\*){3}", "<md-strong-it>\\1</md-strong-it>", text)
    text = re.sub(r"(?:(?![\\])\*){2}([^\*\n]+)(?:(?![\\])\*){2}", "<md-strong>\\1</md-strong>", text)
    text = re.sub(r"(?:(?![\\])\*){1}([^\*\n]+)(?:(?![\\])\*){1}", "<md-it>\\1</md-it>", text)

    # we replace Markdown links
    text = re.sub(r"(?!\\)\[([^\]]+?)(?!\\)\](?!\\)\(([^\)]+?)(?!\\)\)",serialize_md_link
        , text)
    text = re.sub(r"(?!\\)\[([^\]]+?)(?!\\)\]",
        serialize_md_link, text)

    return text


import os
import re
import sys
import logging
import traceback

from .helpers.hash import hash
from .helpers.translate import translate, FileCache
from .helpers.languages import get_source_and_target_languages
from .helpers.serialize import serialize_text, deserialize_text

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("")

def parse_into_blocks(content):
    """
    - We parse Markdown into individual lines
    - If we encounter special ignoretags, we ignore everything inside them
    """
    lines = content.split("\n")
    blocks = []
    ignore_text = ""
    ignore = False
    for line in lines:
        if (not ignore) and re.match(r"^\s*<\!--translate:ignore-->\s*$", line):
            ignore = True
            continue
        elif ignore and re.match(r"<\!--translate:ignore-->\s*$", line):
            blocks.append({'type': 'ignore', 'text' : ignore_text})
            ignore = False
            ignore_text = ""
            continue
        # we ignore everything inside the 'ignore' block
        if ignore:
            ignore_text += line + "\n"
            continue
        blocks.append({'type': 'text', 'text': line})
    return blocks

def translate_file(token, source_path, destination_path, source_language, target_language, clean=False):
    count = 0

    with open(source_path) as input_file:
        content = input_file.read()

    cache = FileCache(source_path+".trans")

    blocks = parse_into_blocks(content)

    for i, block in enumerate(blocks):
        if block['type'] == 'text':
            source_text = serialize_text(block['text'])
            existing_translation = cache.get(source_text, target_language, source_language=source_language)
            if existing_translation is not None:
                # we already have translated this block
                block['translation'] = existing_translation
                continue
            count += len(source_text)
            translation = deserialize_text(translate(source_text, source_language, target_language, token))
            cache.set(source_text, target_language, translation, source_language=source_language)
            block['translation'] = translation

    os.makedirs(os.path.dirname(destination_path), exist_ok=True)

    with open(destination_path, "w") as output_file:
        for block in blocks:
            if block['type'] == 'text':
                output_file.write(block['translation']+"\n")
            elif block['type'] == 'ignore':
                output_file.write(block['text']+"\n")

    if clean:
        cache.clean()

    return count

def translate_markdown(token, src_path, clean=False, match=None):
    match_path = os.path.abspth(match) if match is not None else None
    source_language, target_languages = get_source_and_target_languages(src_path)

    logger.info(f"Translating from '{source_language}' to '{', '.join(target_languages)}'...")
    language_src_path = os.path.join(src_path, source_language)
    for root, dirs, files in os.walk(language_src_path):
        for filename in files:
            if filename.endswith(".md"):
                source_path = os.path.join(root, filename)
                if match_path is not None and match_path != source_path:
                    continue
                elif match_path is not None:
                    print(f"Matched: '{source_path}'")
                for target_language in target_languages:
                    destination_path = os.path.join(src_path, target_language,
                        os.path.relpath(source_path, language_src_path))
                    count = translate_file(token, source_path, destination_path, source_language, target_language, clean=clean)
                    if count:
                        print(f"Translated {count} characters in file {source_path}")



from .yaml import yaml
import os
import sys

def get_source_and_target_languages(src_path, exclude_source=True):

    args = sys.argv[1:]

    site_path = os.path.join(src_path, "site-i18n.yml")

    with open(site_path) as input_file:
        config = yaml.load(input_file.read(), Loader=yaml.FullLoader)

    source_language = 'en'
    if '--tl' in args:
        for i, k in enumerate(args):
            if k == '--tl':
                target_languages = [tl.strip() for tl in args[i+1].split(",")]
                break
    else:
        target_languages = [tl for tl in config['languages'].keys() if not tl.startswith('$')]

    if '--sl' in args:
        for i, k in enumerate(args):
            if k == '--sl':
                source_language = args[i+1].split()
                break
    else:
        source_language = config['i18n']['source_language']

    return source_language, [tl for tl in target_languages if not tl == source_language or not exclude_source]



class BaseBuilder(object):

    """
    A builder generates content for the site in two phases:

    * Indexing (to generate links and structures that can be used by other builders)
    * Building (where the actual content is being generated)
    * Post-Processing (where additional work can be performed)
    """

    def index(self, params, language):
        return {}

    def build(self):
        pass

    def postprocess(self):
        pass

    def __init__(self, site):
        self.site = site
        self.providers = {}
        self.addons = {}

    



class BaseProcessor(object):

    def translate(self, key, *args, **kwargs):
        return self.site.translate(self.language, key, *args, **kwargs)

    def file(self, filename):
        return self.site.request('static_file', filename)

    def source_path(self, filename):
        return self.site.request('source_path', filename)

    def href(self, href, language=None):
        if language is None:
            language = self.language
        return self.site.href(language, href)

    def full_href(self, href, language=None):
        if language is None:
            language = self.language
        return self.site.full_href(language, href)

    def __init__(self, site, params, language):
        self.plaintext = False
        self.site = site
        self.params = params
        self.language = language
    


from beam import Site
from beam.config import load_config

import logging
import yaml
import click
import os

logger = logging.getLogger(__name__)

@click.command("config")
@click.option("--site", default=None)
def config(site):
    if site is None:
        for site in ['site.yml', 'src/site.yml']:
            if os.path.exists(site):
                break
    if not os.path.exists(site):
        logger.error("Site configuration not found.")
        return -1
    config = load_config(site)
    click.echo(yaml.dump(config))




import logging
import click

import beam.i18n as i18n_module

logger = logging.getLogger(__name__)

@click.group("i18n")
def i18n():
    pass

@i18n.command()
@click.option('-c', '--clean', default=False, is_flag=True)
@click.argument("token")
@click.argument("src_path")
def translate_config(token, src_path, clean=False):
    i18n_module.translate_config(token, src_path, clean=clean)

@i18n.command()
@click.option('-c', '--clean', default=False, is_flag=True)
@click.argument("token")
@click.argument("src_path")
def translate_markdown(token, src_path, clean=False):
    i18n_module.translate_markdown(token, src_path, clean=clean)

@i18n.command()
@click.option('-c', '--clean', default=False, is_flag=True)
@click.argument("token")
@click.argument("src_path")
def translate_strings(token, src_path, clean=False):
    i18n_module.translate_strings(token, src_path, clean=clean)

@i18n.command()
@click.option('-c', '--clean', default=False, is_flag=True)
@click.argument("token")
@click.argument("src_path")
def translate(token, src_path, clean=False):
    i18n_module.translate_markdown(token, src_path, clean=clean)
    i18n_module.translate_strings(token, src_path, clean=clean)
    i18n_module.translate_config(token, src_path, clean=clean)



from .base import BaseBuilder
import logging
import traceback

logger = logging.getLogger(__name__)

class PagesBuilder(BaseBuilder):

    def __init__(self, site):
        super().__init__(site)
        self.pages_by_language = {}

    def index(self, params, language):
        pages = self.parse_pages(params.get('pages', []), language)
        self.pages_by_language[language] = pages
        links, link_attrs = self.create_links(pages, language)
        return {
            'links' : links,
            'link_attrs' : link_attrs,
            'vars' : {
                'pages' : pages
            }
        }

    def create_links(self, pages, language):
        links = {}
        link_attrs = {}
        for page in pages:
            if not 'src' in page:
                continue
            names = [page['name']]
            if page.get('index'):
                names.append('')
                # this is the index page
                if '' in links:
                    logger.warning("Multiple index pages defined.")
            for name in names:
                links[name] = page['link']
                link_attrs[name] = page.get('attrs', {})
        return links, link_attrs

    def build(self):
        for language, pages in self.pages_by_language.items():
            for page in pages:
                if not 'src' in page:
                    continue
                self.build_page(page, language)

    def build_page(self, page, language):
        vars = {
            'page' : page,
            'name' : page['name'],
        }
        cp = page
        titles = []
        while cp is not None:
            if 'title' in cp:
                titles += [str(cp['title'])]
            cp = cp.get('parent')
        if titles:
            vars['title'] = ' - '.join(titles[::-1])
        input = self.site.load(page['src'])
        try:
            output = self.site.process(input, page, vars, language)
        except:
            logger.error("An error occured when processing page '{}' in language '{}'"\
                .format(page['name'], language))
            logger.error(traceback.format_exc())
            return
        self.site.write(output, page['dst'])

    def flatten_pages(self, pages):
        """
        Flattens the page hierarchy into a single list, replacing the
        names to reflect the page structure.
        """
        flat_pages = []
        def add_pages(pages, prefix):
            for page in pages:
                new_page = page.copy()
                new_page['children'] = []
                if prefix:
                    new_page['name'] = '.'.join(prefix+[new_page['name']])
                flat_pages.append(new_page)
                if 'children' in page:
                    add_pages(page['children'], prefix+[page['name']])
        add_pages(pages, [])
        return flat_pages 

    def parse_pages(self, pages, language):
        flat_pages = self.flatten_pages(pages)
        pages = self.site.parse_objs(flat_pages, language)
        page_index = {page['name'] : page for page in pages}
        new_slugs = {}
        #we parse child pages and generate appropriate links
        for page in pages:
            components = page['name'].split('.')
            if len(components) > 1:
                full_slug = []
                for i in range(1, len(components)+1):
                    name = '.'.join(components[:i])
                    if name in page_index:
                        page_slug = page_index[name]['slug']
                        if page_slug.startswith('/'):
                            full_slug = [page_slug[1:]]
                        else:
                            full_slug.append(page_slug)
                new_slugs[page['name']] = '/'.join(full_slug)
                page['level'] = len(components)-1
                parent = '.'.join(components[:-1])
                if parent in page_index:
                    parent_page = page_index[parent]
                    page['parent'] = parent_page
                    if not 'children' in parent_page:
                        parent_page['children'] = []
                    parent_page['children'].append(page)
                else:
                    logger.warning("No parent found for page {}".format(page['name']))
            else:
                page['level'] = 0
        #we update the slugs for the child pages
        for name, slug in new_slugs.items():
            page = page_index[name]
            extension = page.get('extension', 'html')
            page['slug'] = slug
            page['dst'] = self.site.get_dst(slug, language, extension=extension)
            page['link'] = self.site.get_link_dst(slug, language, extension=extension)
        return pages



import yaml
# this forces pyyaml to produce cleaner looking strings
# https://stackoverflow.com/questions/8640959/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data
def str_presenter(dumper, data):
    if len(data) > 60:  # check for long strings
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='>')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)

yaml.add_representer(str, str_presenter)




import os
import requests
from .yaml import yaml
from .hash import hash

import logging

logger = logging.getLogger(__name__)

class FileCache:

    def __init__(self, filename):
        self.filename = filename
        self.used = set()
        self.read()

    def read(self):
        if os.path.exists(self.filename):
            with open(self.filename) as input_file:
                self.tr = yaml.load(input_file.read(), Loader=yaml.BaseLoader)
        else:
            self.tr = {}

    def write(self):
        with open(self.filename, 'w') as output_file:
            output_file.write(yaml.dump(self.tr, indent=2, sort_keys=True, width=60))

    def get(self, text, language, source_language=None):
        hh = hash(text)
        self.used.add(f"{hh}")
        if not hh in self.tr or not language in self.tr[hh]:
            return None
        if source_language:
            if not self.has(text, source_language):
                self.set(text, source_language, text)
        return self.tr[hh][language]

    def has(self, text, language):
        hh = hash(text)
        return hh in self.tr and language in self.tr[hh]

    def set(self, text, language, translation, source_language=None):
        hh = hash(text)
        if not hh in self.tr:
            self.tr[hh] = {}
        self.tr[hh][language] = translation
        if source_language:
            self.tr[hh][source_language] = text
        self.write()

    def clean(self):
        for hh, hash_values in list(self.tr.items()):
            if not f"{hh}" in self.used:
                del self.tr[hh]
        self.write()

def cached_translate(text, source_language, target_language, cache, token):
    existing_translation = cache.get(text, target_language, source_language=source_language)
    if existing_translation:
        return existing_translation
    translation = translate(text, source_language, target_language, token)
    cache.set(text, target_language, translation, source_language=source_language)
    return translation

def translate(text, source_language, target_language, token):
    logger.info(f"Translating: '{text}'")
    response = requests.post("https://api.deepl.com/v2/translate", data={
        "auth_key": token,
        "text": text,
        "source_lang": source_language,
        "target_lang": target_language,
        "preserve_formatting": "0",
        "formality": "more",
        "ignore_tags": "ignore,code",
        "tag_handling": "xml",
        })
    if response.status_code != 200:
        logger.error(response.json())
        response.raise_for_status()
    translation = response.json()["translations"][0]["text"]
    logger.info(f"Translation: '{translation}'")
    return translation



from .settings import SETTINGS
from .config import update
from urllib.parse import urlparse
from collections import defaultdict
import importlib
import logging
import copy
import os
import re

from beam.config import load_config

logger = logging.getLogger(__name__)


class Site(object):

    """
    Describes a site.
    """

    def __init__(self, config):
        self.config = copy.deepcopy(config)
        self._original_config = config
        self.settings = {
            'processors' : SETTINGS['processors'].copy(),
            'loaders' : SETTINGS['loaders'].copy(),
            'builders' : SETTINGS['builders'].copy(),
        }
        self._translations = None
        self.process_config()

    @property
    def title(self):
        return self.config.get('title', '')

    @property
    def subtitle(self):
        return self.config.get('subtitle', '')

    @property
    def src_path(self):
        return self.config.get('src-path', 'src')

    @property
    def static_path(self):
        return self.config.get('static-path', 'static')

    @property
    def build_path(self):
        return self.config.get('build-path', 'build')

    @property
    def path(self):
        return self.config.get('path','/')

    @property
    def translations(self):

        def convert_keys(d):
            """
            We convert all keys to strings, which makes it easier to e.g. use
            numbers without explicitly wrapping them in quotes in the YAML files.
            """
            return {
                str(key) : convert_keys(value) if isinstance(value, dict) else value
                for key, value in d.items()
            }

        if self._translations is not None:
            return self._translations

        translations = {}

        if 'translations' in self.config:
            for key, trs in self.config['translations'].items():
                if not key in translations:
                    translations[key] = {}
                translations[key].update(trs)
        self._translations = convert_keys(translations)
        return translations

    def process_config(self):
        if '$all' in self.config.get('languages', {}):
            all_params = self.config['languages']['$all']
            del self.config['languages']['$all']
            for language, params in self.config['languages'].items():
                update(params, all_params)
        if 'builders' in self.config:
            self.settings['builders'].extend(self.config['builders'])

    def translate(self, language, key, fallback=None, *args, **kwargs):
        translations = self.translations
        if not isinstance(key, (list, tuple)):
            kc = key.split('.')
        else:
            kc = key
        cv = translations
        for elem in kc:
            if not elem in cv:
                if fallback:
                    return self.translate(language, fallback, *args, **kwargs)
                return "[no translation for key {}]".format(key)
            cv = cv[elem]
        if not language in cv or not cv[language]:
            if fallback:
                return self.translate(language, fallback, *args, **kwargs)
            return "[no translation for language {} and key {}]".format(language, key)
        text = cv[language].format(*args, **kwargs)
        #  if there are <tr-snip> tags, we only return the text within them.
        if re.match(r".*<tr-snip>", text):
            snippets = []
            for m in re.finditer(r"<tr-snip>([^<]+?)</tr-snip>", text):
                snippets.append(m.group(1).strip())
            return " ".join(snippets)
        return text


    def get_language_prefix(self, language):
        return self.config['languages'][language].get('prefix', language)

    def get_language_link_prefix(self, language):
        return self.config['languages'][language].get('link-prefix', self.get_language_prefix(language))

    def get_src_path(self, path):
        return os.path.abspath(os.path.join(self.src_path, path))

    def get_build_path(self, path):
        return os.path.abspath(os.path.join(self.build_path, path))

    def get_link_dst(self, slug, language, prefix='', extension='html'):
        suffix = ''
        if extension:
            suffix = '.{}'.format(extension)
        return os.path.join(self.get_language_link_prefix(language), prefix, slug)+suffix

    def get_dst(self, slug, language, prefix='', extension='html'):
        suffix = ''
        if extension:
            suffix = '.{}'.format(extension)
        return os.path.join(self.get_language_prefix(language), prefix, slug)+suffix

    def parse_objs(self, objs, language, prefix=''):
        parsed_objs = []
        for obj in objs:
            obj = obj.copy()
            parsed_objs.append(obj)
            if not 'src' in obj:
                #this is just a category page without a source
                continue
            extension = obj.get('extension', 'html')
            if not 'slug' in obj:
                obj['slug'] = ''.join(os.path.basename(obj['src']).split('.')[:-1])
            if not 'dst' in obj:
                obj['dst'] = self.get_dst(obj['slug'], language, prefix, extension=extension)
            if not 'link' in obj:
                obj['link'] = self.get_link_dst(obj['slug'], language, prefix, extension=extension)
            if obj['src'].find('://') == -1:
                obj['src'] = 'file://{}'.format(obj['src'])
            #if not type is given, we use the extension to determine it
            if not 'type' in obj:
                s = obj['src'].split('.')
                if len(s) < 2:
                    raise ValueError
                obj['type'] = s[-1]
        return parsed_objs

    def write(self, content, path):
        logger.info("Writing content to {}".format(path))
        full_path = self.get_build_path(path)
        dirname = os.path.dirname(full_path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(full_path, 'w') as output_file:
            output_file.write(content)

    def url(self, language):
        return self.config['languages'][language].get('url', self.config.get('url', ''))

    def full_href(self, language, url):
        href = self.href(language, url)
        return '{}{}'.format(self.url(language), href)

    def href(self, language, url):
        link = self.get_link(language, url)
        if link:
            hide_extensions = self.config.get('hide-href-extensions', [])
            for extension in hide_extensions:
                # we do not remove the entire link as that breaks it
                if link.endswith(extension) and extension != link:
                    link = link[:-len(extension)]
        else:
            logger.warning("No link for name {} and language {} found".format(url, language))
        return link

    def scss(self, filename):
        return filename

    def load(self, src):
        o = urlparse(src)
        for loader_params in self.settings['loaders']:
            if loader_params['scheme'] == o.scheme:
                break
        else:
            raise TypeError("No loader for scheme: {}".format(o.scheme))
        loader = loader_params['loader'](self)
        path = src[len(o.scheme)+3:]
        return loader.load(path)

    def process(self, input, params, vars, language):
        for processor_params in self.settings['processors']:
            if params['type'] == processor_params['type']:
                break
        else:
            raise TypeError("No processor for file type: {}".format(filename))
        output = input
        full_vars = {
            'params' : params,
        }
        #first we include generic context
        full_vars.update(self.config.get('context', {}))
        #then language-specific context
        full_vars.update(self.config['languages'][language].get('context', {}))
        #then common definitions
        full_vars.update({
            'language' : self.config['languages'][language],
            'languages' : self.config['languages'],
            'lang' : language,
            'site' : self,
        })
        #then builder-specific variables
        full_vars.update(self.vars[language])
        #and finally variables passed as parameters
        full_vars.update(vars)
        for processor_cls in processor_params['processors']:
            processor = processor_cls(self, params, language)
            output = processor.process(output, full_vars)
        return output

    def get_filename(self, language, name):
        return self.links[language][name]

    def get_link(self, language, name):
        try:
            return '{}{}'.format(self.path, self.get_filename(language, name))
        except KeyError:
            return None

    def request(self, name, *args, **kwargs):
        if not name in self.providers:
            raise ValueError("No provider of type {} found!".format(name))
        return self.providers[name](*args, **kwargs)

    def init_builders(self):
        self.links = {}
        self.link_attrs = {}
        self.vars = {}
        self.providers = {}
        self.addons = defaultdict(list)
        self.files = []
        self.builders = []

        for builder_config in self.settings['builders']:
            logging.info("Initializing builder {}...".format(builder_config['name']))
            builder_class = builder_config['builder']
            if isinstance(builder_class, str):
                components = builder_class.split('.')
                builder_module = '.'.join(components[:-1])
                builder_class_str = components[-1]
                try:
                    module = importlib.import_module(builder_module)
                    builder_class = getattr(module, builder_class_str)
                except ImportError:
                    raise
            builder = builder_class(self)
            self.providers.update(builder.providers)
            if builder.addons:
                for key, addon in builder.addons.items():
                    self.addons[key].append(addon)
            self.builders.append(builder)

    def build(self):

        self.init_builders()

        for language, params in self.config.get('languages', {}).items():
            self.links[language] = {}
            self.link_attrs[language] = {}
            self.vars[language] = {}
            for builder in self.builders:
                params['name'] = language
                #here the builders create links and other structures
                result = builder.index(params, language)
                self.links[language].update(result.get('links', {}))
                self.link_attrs[language].update(result.get('link_attrs', {}))
                self.vars[language].update(result.get('vars', {}))

        for builder in self.builders:
            #now the builder "build" their components
            builder.build()

        for builder in self.builders:
            #now builders can do some post-processing
            builder.postprocess()



import yaml
import json
import os

def update(d, ud):
    for key, value in ud.items():
        overwrite = False
        if key.endswith('!'):
            key = key[:-1]
            overwrite = True
        if key not in d:
            d[key] = value
        elif isinstance(value, dict):
            if overwrite:
                d[key] = value
            else:
                update(d[key], value)
        elif isinstance(value, list) and isinstance(d[key], list):
            if overwrite:
                d[key] = value
            else:
                d[key] += value
        elif overwrite:
            d[key] = value

def load_include(value, include_path, with_data=True):
    path = os.path.abspath(os.path.join(os.path.dirname(include_path[-1]), value))
    if path in include_path:
        raise ValueError("Recursive import of {} (path: {})"\
            .format(path, '->'\
            .join(['"{}"'\
            .format(s) for s in include_path])))
    return load_config(path, include_path=include_path+[path], with_data=with_data)

def load_includes(config, include_path, with_data=True):
    if isinstance(config, dict):

        d = config.copy()

        for key, value in d.items():
            d[key] = load_includes(value, include_path=include_path, with_data=with_data)

        if '$include' in d:
            if not with_data and d.get("$data"):
                return d
            if d.get('$as-list'):
                nds = []
                is_list = True
            else:
                is_list = False
                nds = d.copy()
                del nds['$include']

            includes = d['$include']

            if not isinstance(includes, list):
                includes = [includes]
            for include in includes:
                nd = load_include(include, include_path, with_data=with_data)
                if nd is None:
                    continue
                if is_list:
                    if not isinstance(nd, (list, tuple)):
                        raise ValueError("expected a list or tuple")
                    nds.extend(nd)
                else:
                    if not isinstance(nd, dict):
                        raise ValueError("expected a dictionary")
                    update(nds, nd)
            return nds
        return d
    elif isinstance(config, (list, tuple)):
        l = []
        for c in config:
            result = load_includes(c, include_path=include_path, with_data=with_data)
            if isinstance(c, dict) and '$include' in c and isinstance(result, list):
                l.extend(result)
            else:
                l.append(result)
        return l
    return config
            

def load_config(filename, include_path=None, with_data=True):
    if include_path is None:
        include_path = [os.path.abspath(filename)]
    with open(filename) as input_file:
        if filename.endswith(".json"):
            # this is a JSON file
            config = json.load(input_file)
        else:
            config = yaml.load(input_file.read(),Loader=yaml.FullLoader)
    return load_includes(config, include_path=include_path, with_data=with_data)



import os
import re
import sys
import logging
import traceback

from beam.config import load_config
from .helpers.hash import hash
from .helpers.yaml import yaml
from .helpers.languages import get_source_and_target_languages
from .helpers.serialize import serialize_text, deserialize_text
from .helpers.translate import cached_translate, FileCache

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("")

class Skip(BaseException):
    pass

def transform_data(data, context, filters, parent_key=None):
    if isinstance(data, dict):
        for r in [0,1,2]:
            nd = {}
            for key, value in data.items():
                sv = key.split("|")
                fns = sv[1:]
                relevant_fns = []
                pipes = 0
                new_key = key
                for fn in fns:
                    if not fn: # this matched a pipe
                        pipes += 1
                        continue
                    elif pipes != r:
                        pipes = 0
                        continue
                    pipes = 0
                    relevant_fns.append(fn)
                if len(relevant_fns) == len(fns) or r == 2:
                    new_key = sv[0]
                for fn in relevant_fns:
                    if not fn in filters:
                        logger.warning(f"Skipping undefined filter {fn}...")
                        continue
                    ff = filters[fn]
                    try:
                        if isinstance(value, list):
                            new_value = []
                            for element in value:
                                new_value.append(ff(new_key, element, data, context))
                            value = new_value
                        else:
                            value = ff(new_key, value, data, context)
                    except Skip:
                        raise
                    except:
                        logger.error(f"Error processing key {key} with parent key {parent_key}")
                        raise
                nd[new_key] = value
            data = nd
        nnd = {}
        for key, value in nd.items():
            try:
                full_key = f"{parent_key}.{key}" if parent_key else key
                nnd[key] = transform_data(value, context, filters, parent_key=full_key)
            except Skip:
                pass
        return nnd
    elif isinstance(data, list):
        nd = []
        for i, element in enumerate(data):
            try:
                nd.append(transform_data(element, context, filters, parent_key=f'{i}' if not parent_key else f'{parent_key}.{i}'))
            except Skip:
                pass
        return nd
    else:
        return data

def exists(key, value, data, context):
    path = os.path.join(context["pwd"], value)
    if not os.path.exists(path):
        logger.debug(f"Skipping {path}")
        raise Skip
    return value

keymap = {
    'ä' : 'ae',
    'ö' : 'oe',
    'ü' : 'ue',
    'ß' : 'ss',    
}

def slugify(key, value, data, context):
    title = data["title"].lower().replace(" ", "-")
    for sk, dk in keymap.items():
        title = title.replace(sk, dk)
    title = re.sub(r"[^a-z\-]", "", title)
    if not title: # for 'exotic' languages like Chinese...
        return re.sub(r"[^a-z\-]", "", data["name"].lower().replace(" ", "-"))
    return title

def translate_file(token, site_path, destination_path, source_language, target_language, cache):
    count = 0
    source_data = load_config(site_path, with_data=False)

    filters = {
        't' : lambda key, value, data, context: value if source_language == target_language else deserialize_text(cached_translate(serialize_text(value), source_language, target_language, cache, token)),
        'f' : lambda key, value, data, context: value.format(**context),
        'exists?': exists,
        'id' : lambda key, value, data, context: value,
        'ld' : lambda key, value, data, context: value.get(context['target_language'], value['default']),
        'slugify': slugify,
    }

    """
    We iterate through the file, parsing the modifiers and generating a new
    config from them...
    """
    transformed_data = transform_data(source_data, {
        "source_language" : source_language,
        "target_language" : target_language,
        "pwd" : os.path.dirname(site_path),
        }, filters)
    os.makedirs(os.path.dirname(destination_path), exist_ok=True)
    with open(destination_path, "w") as output_file:
        output_file.write("# Warning, this is an auto-generated file, do not modify!\n" +
                          f"# Modify this file instead: {os.path.relpath(site_path, os.path.dirname(destination_path))}\n")
        output_file.write(yaml.dump(transformed_data, indent=2, sort_keys=True, width=40))


def translate_config(token, src_path, site_name="site-all.yml", clean=False):

    source_language, target_languages = get_source_and_target_languages(src_path, exclude_source=False)

    site_path = os.path.join(src_path, site_name)
    if not os.path.exists(site_path):
        logger.fatal("Master file does not exist!")
        exit(-1)
    logger.info(f"Translating file '{site_path}' from '{source_language}' to '{', '.join(target_languages)}'...")

    cache = FileCache(site_path+".trans")

    for target_language in target_languages:
        destination_path = os.path.join(src_path, f"site-{target_language}.yml")
        count = translate_file(token, site_path, destination_path, source_language, target_language, cache)
        if count:
            logger.info(f"Translated {count} characters in file {site_path}")
    if clean:
        cache.clean()



from .base import BaseBuilder

import datetime
import math
import os

class BlogBuilder(BaseBuilder):

    def __init__(self, site):
        super().__init__(site)
        self.blog_pages_by_language = {}

    def index(self, params, language):
        articles = self.parse_articles(params.get('articles', []), language)
        blog_pages = self.paginate_articles(self.sort_articles(articles))
        self.blog_pages_by_language[language] = blog_pages
        links = self.create_links(blog_pages, articles, language)
        return {
            'links' : links,
            'vars' : {
                'articles' : articles
            },
        }

    def get_index_slug(self, language, i):
        return os.path.join(
            self.get_blog_prefix(language),
            'index{}'.format('-{}'.format(i) if i != 0 else '')
        )

    def create_links(self, blog_pages, articles, language):
        links = {}
        for i, blog_page in enumerate(blog_pages):
            slug = self.get_index_slug(language, i)
            link = self.site.get_link_dst(slug, language)
            links['blog-{}'.format(i)] = link
            if i == 0:
                links['blog'] = link

            for article in articles:
                links['blog-{}'.format(article['name'])] = article['dst']
        return links

    def build(self):
        for language, blog_pages in self.blog_pages_by_language.items():
            self.build_blog(blog_pages, language)

    def sort_articles(self, articles):
        sorted_articles = list(reversed(sorted(articles, key=lambda x : x.get('date',x.get('title')))))
        return sorted_articles

    def paginate_articles(self, articles):
        app = self.site.config.get('articles-per-page', 10)
        return [articles[i*app:(i+1)*app] for i in range(math.ceil(len(articles)/app))]

    def build_blog(self, blog_pages, language):
        """
        Build the indexes, meta-pages and articles.
        """
        for i, blog_page in enumerate(blog_pages):
            self.build_index_site(i, len(blog_pages), blog_page, language)
            for article in blog_page:
                self.build_article(article, i, language)

    def get_blog_prefix(self, language):
        return self.site.config['languages'][language].get('blog-path',
            self.site.config.get('blog-path', 'blog'))

    def build_index_site(self, i, n, blog_page, language):
        input = "{% extends('index.html') %}"
        vars = {
            'name' : 'blog-{}'.format(i),
            'blog_page' : blog_page,
            'i' : i,
            'n' : n,
        }
        slug = self.get_index_slug(language, i)
        dst = self.site.get_dst(slug, language)
        output = self.site.process(input, {'type' : 'html'}, vars, language)
        self.site.write(output, dst)

    def build_article(self, article, page, language):
        """
        Build an individual blog article.
        """
        vars = {
            'article' : article,
            'name' : 'blog-{}'.format(article['name']),
            'slug' : article.get('slug', article['name']),
            'blog_page' : page,
            'index_link' : self.site.get_link(language, 'blog-{}'.format(page)),
        }
        input = self.site.load(article['src'])
        output = self.site.process(input, article, vars, language)
        self.site.write(output, article['dst'])

    def parse_articles(self, articles, language):
        parsed_articles = self.site.parse_objs(articles, language, prefix=self.get_blog_prefix(language))
        for article in parsed_articles:
            date_format = self.site.config['languages'][language].get('date-format', '%Y-%m-%d')
            article['date'] = datetime.datetime.strptime(article['date'], '%Y-%m-%d %H:%M')
            article['date-str'] = article['date'].strftime(date_format)
        return parsed_articles




class Context(object):

    def __init__(self, site):
        self.site = site
        self._d = {}

    def __getitem__(self, item):
        return self._d[item]

    def __setitem__(self, item, value):
        self._d[item] = value

    def __delitem__(self, item):
        del self.d_[item]



from .base import BaseProcessor

import markdown2

template = """
{{% extends('{template}') %}}
{{% block article %}}
{content}
{{% endblock %}}
"""

class MarkdownProcessor(BaseProcessor):

    def process(self, input, vars):
        #we add an offset if requested
        if self.params.get('h-offset') is not None:
            offset = self.params['h-offset']
            lines = input.split('\n')
            input = ""
            for line in lines:
                if line.startswith('#'):
                    line = '#'*offset+line
                input += line + "\n"
        result = markdown2.markdown(input, extras=['footnotes','fenced-code-blocks'])
        if self.params.get('bare'):
            return result
        template_filename = self.params.get('template', self.site.config.get('markdown', {}).get('template', 'markdown.html'))
        template_result = template.format(template=template_filename, content=result)
        return template_result


from distutils.core import setup
from setuptools import find_packages

setup(
    name='beam-up',
    python_requires='>=3',
    version='1.0.20',
    author='Andreas Dewes - 7scientists',
    author_email='andreas.dewes@7scientists.com',
    license='MIT',
    url='https://github.com/adewes/beam',
    packages=find_packages(),
    package_data={'': ['*.ini']},
    include_package_data=True,
    install_requires=['requests', 'click', 'pyyaml', 'markdown2', 'jinja2', 'pygments', 'Pillow'],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'beam = beam.cli.main:beam'
        ]
    },
    description='A simple yet powerful static site generator aimed at product pages (not blogs).',
    long_description="""A simple yet powerful static site generator for Python.

Templates (HTML, Markdown, ...) go in, pages come out.
"""
)



from .base import BaseLoader

class FileLoader(BaseLoader):

    def __init__(self, site):
        super().__init__()
        self.site = site

    def load(self, path):
        abs_path = self.site.get_src_path(path)
        with open(abs_path) as input_file:
            return input_file.read()



from .processors import JinjaProcessor, MarkdownProcessor, PlainProcessor
from .builders import BlogBuilder, PagesBuilder, StaticFilesBuilder, SitemapBuilder
from .loaders import FileLoader

SETTINGS = {
    'defaults' : {
        'src-dir' : 'src',
        'build-dir' : 'build',
    },
    'loaders' : [
        {
            'scheme' : 'file',
            'loader' : FileLoader
        }
    ],
    'builders' : [
        {
            'builder' : BlogBuilder,
            'name' : 'blog'
        },
        {
            'builder' : PagesBuilder,
            'name' : 'pages'
        },
        {
            'builder' : StaticFilesBuilder,
            'providers' : ['static_file'], 
            'name' : 'static_files'
        },
        {
            'builder' : SitemapBuilder,
            'name' : 'sitemap'
        },
    ],
    'processors' : [
        {
            'type' : 'html',
            'processors' : [JinjaProcessor],
            'name' : 'HTML'
        },
        {
            'type' : 'md',
            'processors' : [MarkdownProcessor, JinjaProcessor],
            'name' : 'Markdown'
        },
        {
            'type' : 'plain',
            'processors' : [PlainProcessor],
            'name' : 'Plain'
        }
    ],
}


from .base import BaseProcessor

from collections import defaultdict
from urllib.parse import quote

import logging
import markdown2
import textwrap
import os
import re

logger = logging.getLogger(__name__)

with_pil = False

try:
    from PIL import Image, UnidentifiedImageError
    with_pil = True
except ImportError:
    pass

from jinja2 import Environment, FileSystemLoader, ChoiceLoader, DictLoader

try:
    from pygments import highlight
    from pygments.styles import get_style_by_name
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import HtmlFormatter
    with_pygments = True
except ImportError:
    with_pygments = False

try:
    from bs4 import BeautifulSoup as bs
    with_bs = True
except ImportError:
    with_bs = False

class JinjaProcessor(BaseProcessor):

    def highlight_styles(self, code, style_name=None):
        if style_name is None:
            style_name = self.site.config.get('pygments', {}).get('style', 'monokai')
        style = get_style_by_name(style_name)
        return '<style type="text/css">{}</style>'.format(HtmlFormatter(style=style).get_style_defs('.highlight .{}'.format(style_name)))

    def highlight(self, code, language='python', style_name=None, strip=True, deindent=True):
        if style_name is None:
            style_name = self.site.config.get('pygments', {}).get('style', 'monokai')
        lexer = get_lexer_by_name(language)
        style = get_style_by_name(style_name)
        if deindent:
            code = textwrap.dedent(code)
        if strip:
            code = code.strip()
        return highlight(code, lexer, HtmlFormatter(style=style, cssclass='{}'.format(style_name)))

    def picture(self, filename, prefs=['webp', 'png', 'jpeg'], sizes='50vw', **kwargs):
        file_path = self.file(filename)
        file_dir = os.path.dirname(file_path)
        if not file_path:
            return ""
        source_path = self.source_path(filename)
        filename = os.path.basename(source_path)
        source_dir = os.path.dirname(source_path)
        basename, zoom, ext = re.match(r"^(.*?)(?:@(\d+x))?\.(.*)$", filename).groups()
        extra_args = []
        for k, v in kwargs.items():
            extra_args.append(f'{k}="{v}"')
        extra_args = " ".join(extra_args)
        files = os.listdir(source_dir)
        alternatives = defaultdict(list)
        for fname in files:
            if re.match("{}(?:@\d+x)?\.\w+$".format(re.escape(basename)), fname):
                _, zoom, ext = re.match(r"^(.*?)(?:@(\d+x))?\.(.*)$", fname).groups()
                if zoom is None:
                    zoom = '1x'
                if with_pil:
                    try:
                        # we determine the exact image width to include
                        image = Image.open(os.path.join(source_dir, fname))
                        zoom = '{}w'.format(image.size[0])
                    except UnidentifiedImageError:
                        # we cannot properly read this image, ignore it
                        logger.warning(f"Cannot read image {source_dir}/{fname}, ignoring its size...")
                        pass
                alternatives[ext].append((fname, zoom))

        html_alternatives = ""

        sizestr = ""
        if sizes:
            sizestr = f'sizes="{sizes}"'

        def generate_alternative(ext, ext_alternatives):

            alts = []
            for f, z in ext_alternatives:
                fp = quote(f"{file_dir}/{f}")
                alts.append(f"{fp} {z}")
            joined_alternatives = ", ".join(alts)
            return f"""
    <source type="image/{ext}" srcset="{joined_alternatives}" {sizestr}>
"""

        # we add the preferred formats first
        for pref in prefs:
            if pref in alternatives:
                html_alternatives += generate_alternative(pref, alternatives[pref])
        # then we add the rest
        for ext, ext_alternatives in alternatives.items():
            if not ext in prefs:
                html_alternatives += generate_alternative(ext, ext_alternatives)
        return f"""
<picture>
    {html_alternatives}
    <img src="{file_path}" {extra_args}>
</picture>
"""

    def markdown(self, text):
        result = markdown2.markdown(text, extras=['footnotes', 'fenced-code-blocks'])
        return result

    def translate(self, key, *args, **kwargs):
        result = super().translate(key, *args, **kwargs)
        translate_hint = self.site.config.get('translate-hint')
        if translate_hint:
            return translate_hint.format(value=result, key=key, args=args, kwargs=kwargs)
        return result

    def jinja(self, vars, input, **kwargs):
        context = vars.copy()
        context.update(kwargs)
        return self.process(input, context)

    def get_jinja_env(self, input, vars):
        dict_loader = DictLoader({'input' : input})
        choice_loader = ChoiceLoader([dict_loader, FileSystemLoader('{}/templates'.format(self.site.src_path)), FileSystemLoader(self.site.src_path)])
        env = Environment(loader=choice_loader)

        # we add some useful filters
        env.filters['href'] = self.href
        env.filters['full_href'] = self.full_href
        env.filters['file'] = self.file
        env.filters['markdown'] = self.markdown
        env.filters['picture'] = self.picture
        env.filters['jinja'] = lambda input, **kwargs: self.jinja(vars, input, **kwargs)
        if with_pygments:
            env.filters['highlight'] = self.highlight
            env.filters['highlight_styles'] = self.highlight_styles
        env.filters['translate'] = self.translate
        for filters in self.site.addons['jinja-filters']:
            for name, f in filters:
                env.filters[name] = f
        return env

    def process(self, input, vars):
        env = self.get_jinja_env(input, vars)
        template = env.get_template('input')
        result = template.render(**vars)
        if with_bs and False:
            soup=bs(result, "html.parser")
            return soup.prettify()
        return result


#!/usr/bin/env python3
import click
import logging

from beam.cli import commands

@click.group()
@click.option('-v', '--verbose', count=True)
def beam(verbose):
    #we want to always show errors and warnings
    if verbose == 0:
        verbose = 1
    levels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    level = levels[min(verbose,len(levels)-1)]
    logging.basicConfig(level=level, format='%(message)s')

for command in commands:
    beam.add_command(command)

if __name__ == '__main__':
    beam()